SystemOrganization addCategory: #'CTranslator-Hierarchy'!
SystemOrganization addCategory: #'CTranslator-Translators'!
SystemOrganization addCategory: #'CTranslator-Exceptions'!
SystemOrganization addCategory: #'CTranslator-Programs'!
SystemOrganization addCategory: #'CTranslator-Output'!

RBProgramNodeVisitor subclass: #CMethodTranslator
	instanceVariableNames: 'source class builder'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CMethodTranslator class methodsFor: 'instance creation' stamp: 'Richo 6/28/2011 11:29'!
new
	^ self shouldNotImplement! !

!CMethodTranslator class methodsFor: 'translating' stamp: 'Richo 6/28/2011 11:28'!
translate: selector of: class with: aBuilder
	^ (self withBuilder: aBuilder) translate: selector of: class! !

!CMethodTranslator class methodsFor: 'instance creation' stamp: 'Richo 6/28/2011 11:29'!
withBuilder: aBuilder
	^ self basicNew initializeWith: aBuilder! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptAssignmentNode: anAssignmentNode
	self visitNode: anAssignmentNode variable.
	source nextPutAll: ' = '.
	self visitNode: anAssignmentNode value.
	source nextPut: $;; cr.! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptCascadeNode: aCascadeNode
	aCascadeNode messages do: [:each | self visitNode: each. source nextPut: $;; cr].! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptLiteralNode: aLiteralNode
	| value |
	value := aLiteralNode value.
	value isFloat 
		ifTrue: [source nextPutAll: (value printShowingDecimalPlaces: 12).
			^ self].
	value isString 
		ifTrue: [source nextPut: $"; nextPutAll: value asString; nextPut: $".
			^ self].
	source nextPutAll: value asString! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/21/2011 18:00'!
acceptMessageNode: aMessageNode 
	| method receiverClass |
	receiverClass := CTypeCollector findTypeOf: aMessageNode receiver in: class.	
	method := receiverClass ultimateCompiledMethodAt: aMessageNode selector ifAbsent: [self error: 'Method not found'].

	"Is it a c primitive?"
	(method pragmaAt: #c:)
		ifNotNilDo: [:nxcPrim | source nextPutAll: ((CPrimitiveRule on: nxcPrim arguments first) applyTo: aMessageNode using: self).
			^ self].

	"Should be inlined?"
	(method pragmas anySatisfy: [:each | each keyword = #inlined])
		ifTrue: [self withClass: receiverClass do: [self visitNode: (method methodClass parseTreeFor: aMessageNode selector) body].
			source skip: -2. "Remove the extra semicolon"
			^ self].

	"If we got here, it is a regular method call and we need to signal the MethodToImplement exception."
	(CMethodToImplement
		class: receiverClass
		selector: aMessageNode selector)
			signal.
	self storeRegularMessageSend: aMessageNode! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 21:11'!
acceptMethodNode: methodNode
	self storeMethodDeclaration: methodNode.
	source cr; nextPut: ${; cr.
	methodNode body temporaries do: [:var |  
		source nextPutAll: (CTypeCollector findTypeOf: var in: class) asCType; 
			space; 
			nextPutAll: var name; 
			nextPut: $;; cr].
	self visitNode: methodNode body.
	source nextPut: $}.! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptPragmaNode: aPragmaNode
	"Do nothing"! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptReturnNode: aReturnNode
	source nextPutAll: 'return '.
	self visitNode: aReturnNode value.
	source nextPut: $;; cr.! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptSequenceNode: aSequenceNode
aSequenceNode statements do: [:each | 
	self visitNode: each.
	each isMessage ifTrue: [source nextPut: $;; cr]]! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/22/2011 16:30'!
acceptVariableNode: aVariableNode
	(class instVarNames includes: aVariableNode name)
		ifTrue: [(CInstVarToImplement class: class var: aVariableNode) signal.
			source nextPutAll: 'self.'].
	source nextPutAll: aVariableNode name.! !

!CMethodTranslator methodsFor: 'accessing' stamp: 'Richo 6/28/2011 11:30'!
builder
	^ builder! !

!CMethodTranslator methodsFor: 'initializing' stamp: 'Richo 6/28/2011 11:34'!
initializeWith: aBuilder
	builder := aBuilder.
	self initialize! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/21/2011 14:06'!
storeMethodDeclaration: methodNode 
	| methodName |
	methodName := methodNode selector.
	methodName = #main
		ifTrue: [methodName := CProgramTranslator defaultMainSelector]
		ifFalse: [methodName := methodName asCIdentifier].
	source nextPutAll: (CTypeCollector findTypeOf: methodNode in: class) asCType.
	source nextPutAll: ' ', methodName, '(', class asCType, ' self'.
	methodNode arguments notEmpty
		ifTrue: [source nextPutAll: ', '].
	methodNode arguments 
		do: [:arg | source nextPutAll: (CTypeCollector findTypeOf: arg in: class) asCType, ' ', arg name] 
		separatedBy: [source nextPutAll: ', '].
	source nextPut: $)! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/21/2011 18:00'!
storeRegularMessageSend: aMessageNode
	source nextPutAll: aMessageNode selector asCIdentifier, '('.
	{aMessageNode receiver}, aMessageNode arguments
		do: [:each | self visitNode: each]
		separatedBy: [source nextPutAll: ', '].
	source nextPut: $)! !

!CMethodTranslator methodsFor: 'translating' stamp: 'Richo 6/28/2011 11:22'!
translate: selector of: aClass
	| methodNode |
	source := WriteStream on: ''.
	class := aClass.
	methodNode := (RBParser parseMethod: (class ultimateSourceCodeAt: selector ifAbsent: [self error: 'Source code not found'])).
	self visitNode: methodNode.
	^ self builder newFunctionFor: class node: methodNode source: source contents.! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
visitArguments: aNodeCollection 
	"Do nothing"! !

!CMethodTranslator methodsFor: 'translating' stamp: 'Richo 6/20/2011 20:59'!
withClass: aClass do: aBlock
	"Replaces the class just for running the block, then it goes back to the old class.
	Very useful for translating inlined methods, which needs to use the same source but change the current class"
	| old |
	old := class.
	class := aClass.
	aBlock value.
	class := old! !

!CMethodTranslator methodsFor: 'translating' stamp: 'Richo 6/20/2011 21:12'!
withSource: aWriteStream do: aBlock
	"Replaces the source stream just for running the block, then it goes back to the old source.
	Very useful for CPrimitiveRule, which needs to translate just specific expressions without messing with already translated stuff"
	| oldSource |
	oldSource := source.
	source := aWriteStream.
	aBlock value.
	source := oldSource! !

CMethodTranslator subclass: #CppMethodTranslator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CppMethodTranslator methodsFor: 'as yet unclassified' stamp: 'Richo 6/22/2011 16:38'!
acceptVariableNode: aVariableNode
	(class bindingOf: aVariableNode name)
		ifNotNil: [:association |
			association value isLiteral
				ifTrue: [source nextPutAll: association value asString; space;
							nextPutAll: '/* ', association key, ' */'.
					^ self]].
	(class instVarNames includes: aVariableNode name)
		ifTrue: [(CInstVarToImplement class: class var: aVariableNode) signal].
	source nextPutAll: aVariableNode name.! !

!CppMethodTranslator methodsFor: 'as yet unclassified' stamp: 'Richo 6/22/2011 11:13'!
storeMethodDeclaration: methodNode 
	| methodName |
	methodName := methodNode selector.
	methodName = #main
		ifTrue: [methodName := CProgramTranslator defaultMainSelector]
		ifFalse: [methodName := methodName asCIdentifier].
	source nextPutAll: (CTypeCollector findTypeOf: methodNode in: class) asCType.
	source nextPutAll: ' ', class name, '::', methodName, '('.
	methodNode arguments 
		do: [:arg | source nextPutAll: (CTypeCollector findTypeOf: arg in: class) asCType, ' ', arg name] 
		separatedBy: [source nextPutAll: ', '].
	source nextPut: $)! !

!CppMethodTranslator methodsFor: 'as yet unclassified' stamp: 'Richo 6/22/2011 11:58'!
storeRegularMessageSend: aMessageNode
	source nextPutAll: aMessageNode selector asCIdentifier, '('.
	aMessageNode arguments
		do: [:each | self visitNode: each]
		separatedBy: [source nextPutAll: ', '].
	source nextPut: $)! !

RBProgramNodeVisitor subclass: #CTypeCollector
	instanceVariableNames: 'selfClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CTypeCollector class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 20:59'!
findTypeOf: aProgramNode in: aClass 
	^ self new findTypeOf: aProgramNode in: aClass ! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptArrayNode: anArrayNode 
	self halt! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 21:21'!
acceptBlockNode: aBlockNode 
	^ CBlock! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptCascadeNode: aCascadeNode 
	^ self visitNode: aCascadeNode messages last! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptLiteralArrayNode: aRBLiteralArrayNode 
	self halt! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 21:22'!
acceptLiteralNode: aLiteralNode
	aLiteralNode value isFloat
		ifTrue: [^ CFloat].
	aLiteralNode value isNumber
		ifTrue: [^CLong].
	(#(true false) includes: aLiteralNode value)
		ifTrue: [^ CBoolean].
	aLiteralNode value isString
		ifTrue: [^ CString].
	self halt! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/22/2011 16:28'!
acceptMessageNode: aMessageNode 
	"It looks for the pragma #returnType: in the method being called. If not found, it returns the receiver's class."
	| receiverClass method |
	receiverClass := self findTypeOf: aMessageNode receiver.
	method := receiverClass ultimateCompiledMethodAt: aMessageNode selector ifAbsent: [self error: 'Method not found'].
	
	(method pragmaAt: #returnType:)
		ifNotNilDo: [:pragma | ^ (Smalltalk at: pragma arguments first value first)].

	^ receiverClass! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/22/2011 16:27'!
acceptMethodNode: aMethodNode
	"It looks for the #returnType: pragma to get the return type of the method"
	(aMethodNode pragmas detect: [:each | each selector = #returnType:] ifNone: [nil])
		ifNotNilDo: [:pragma | ^ Smalltalk at: pragma arguments first value first].
	^ CVoid! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptPatternBlockNode: aRBPatternBlockNode 
	self halt! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptPatternWrapperBlockNode: aRBPatternWrapperBlockNode 
	self halt! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptSequenceNode: aSequenceNode 
	^ self visitNode: aSequenceNode statements last! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/23/2011 17:20'!
acceptVariableNode: aVariableNode 
	| methodNode findPragma className |
	methodNode := aVariableNode methodNode.
	
	findPragma := [:selector | "Looks in the pragmas for the given selector and returns the arguments value"
		(methodNode pragmas detect: [:each | each selector = selector] ifNone: [])
				ifNotNilDo: [:pragma | pragma arguments first value]].

	"Is it an argument?"
	(aVariableNode methodNode arguments includes: aVariableNode)
		ifTrue: [(findPragma value: #argTypes:) ifNotNilDo: [:argTypes | 
					^ (self computeTypes: argTypes) at: (methodNode arguments indexOf: aVariableNode)]].

	"Is it a temp var?"
	(aVariableNode temporaryVariables includes: aVariableNode name)
		ifTrue: [(findPragma value: #tempTypes:) ifNotNilDo: [:tempTypes |
					^ (self computeTypes: tempTypes) at: (aVariableNode temporaryVariables indexOf: aVariableNode name)]].

	"Is it an inst var?"
	(selfClass instVarNames includes: aVariableNode name)
		ifTrue: [className := selfClass instanceVariableTypes at: (selfClass instVarNames indexOf: aVariableNode name).
		"In theory, this should never throw an error"
		^ Smalltalk
			at: className
			ifAbsent: [self error: 'Invalid type: ', className]].
	

	self error: 'Variable''s type not found'! !

!CTypeCollector methodsFor: 'utility' stamp: 'Richo 6/23/2011 17:52'!
computeTypes: anArray 
	"It reads the array from method tags and returns an array of c types"
	^ Array streamContents: [:outputStream | | inputStream |
	inputStream := anArray readStream.
	[inputStream atEnd]
		whileFalse: [| next |
			next := Smalltalk at: inputStream next ifAbsent: [self error: 'Invalid type: '].
			inputStream peek = #*
				ifTrue: [inputStream next.
					next := CPointer on: next].
			outputStream nextPut: next]].
		! !

!CTypeCollector methodsFor: 'finding types' stamp: 'Richo 6/20/2011 20:59'!
findTypeOf: aProgramNode
	aProgramNode = (RBVariableNode named: 'self')
		ifTrue: [^ selfClass].
	^ self visitNode: aProgramNode.
! !

!CTypeCollector methodsFor: 'finding types' stamp: 'Richo 6/20/2011 20:59'!
findTypeOf: aProgramNode in: aClass 
	selfClass := aClass.
	^ self findTypeOf: aProgramNode! !

Exception subclass: #CInstVarToImplement
	instanceVariableNames: 'ownerClass variable'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Exceptions'!

!CInstVarToImplement class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 21:05'!
class: aClass var: aVariableNode 
	^ (self basicNew setClass: aClass var: aVariableNode) initialize.! !

!CInstVarToImplement methodsFor: 'exceptionDescription' stamp: 'Richo 6/20/2011 21:05'!
defaultAction
	"No one has handled this error, but now give them a chance to decide how to debug it.  If none handle this either then open debugger (see UnhandedError-defaultAction)"
	UnhandledError signalForException: self! !

!CInstVarToImplement methodsFor: 'accessing' stamp: 'Richo 6/20/2011 21:05'!
ownerClass
	^ ownerClass! !

!CInstVarToImplement methodsFor: 'private' stamp: 'Richo 6/20/2011 21:05'!
setClass: aClass var: aVariableNode
	ownerClass := aClass.
	variable := aVariableNode! !

!CInstVarToImplement methodsFor: 'accessing' stamp: 'Richo 6/20/2011 21:05'!
varName
	^ variable name! !

!CInstVarToImplement methodsFor: 'accessing' stamp: 'Richo 6/21/2011 15:03'!
varType
	^ CTypeCollector findTypeOf: variable in: ownerClass.! !

Exception subclass: #CMethodToImplement
	instanceVariableNames: 'methodClass methodSelector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Exceptions'!

!CMethodToImplement class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 21:04'!
class: aClass selector: aByteSymbol 
	^ (self basicNew setClass: aClass selector: aByteSymbol) initialize! !

!CMethodToImplement methodsFor: 'exceptionDescription' stamp: 'Richo 6/20/2011 21:04'!
defaultAction
	"No one has handled this error, but now give them a chance to decide how to debug it.  If none handle this either then open debugger (see UnhandedError-defaultAction)"
	UnhandledError signalForException: self! !

!CMethodToImplement methodsFor: 'printing' stamp: 'Richo 6/20/2011 21:04'!
description
	"Return a textual description of the exception."
	^ 'To implement: ' , methodClass asString , '>>' , methodSelector printString! !

!CMethodToImplement methodsFor: 'accessing' stamp: 'Richo 6/20/2011 21:04'!
methodClass
	^ methodClass! !

!CMethodToImplement methodsFor: 'accessing' stamp: 'Richo 6/20/2011 21:04'!
methodSelector
	^ methodSelector ! !

!CMethodToImplement methodsFor: 'private' stamp: 'Richo 6/20/2011 21:04'!
setClass: aClass selector: aByteSymbol 
	methodClass := aClass.
	methodSelector := aByteSymbol! !

!String methodsFor: '*CTranslator' stamp: 'Richo 6/28/2011 10:50'!
asCIdentifier
"
'Richo' asCIdentifier.
'Richo es un capo' asCIdentifier.
'foo_bar(234)' asCIdentifier
"

^ self inject: '' into: [:subTotal :next |
	next isAlphaNumeric 
		ifTrue: [subTotal, next asString]
		ifFalse: [subTotal , '_', next asciiValue asString]]! !

Object subclass: #CObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CObject class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^ self name! !

!CObject class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:53'!
definition
	"Answer a String that defines the receiver."
	"Overwritten to show the #instanceVariablesTypes: message in the class definition of CObject subclasses"
	| aStream path |
	self = CObject ifTrue: [^super definition].
	aStream := WriteStream
				on: (String new: 300).
	superclass == nil
		ifTrue: [aStream nextPutAll: 'ProtoObject']
		ifFalse: [path := ''.
			self environment
				scopeFor: superclass name
				from: nil
				envtAndPathIfFound: [:envt :remotePath | path := remotePath].
			aStream nextPutAll: path , superclass name].
	aStream nextPutAll: self kindOfSubclass;
		 store: self name.
	[(self hasTraitComposition
			and: [self traitComposition notEmpty])
		ifTrue: [aStream cr; tab; nextPutAll: 'uses: '; nextPutAll: self traitCompositionString]] on: MessageNotUnderstood do: ["Ignore DNU. Traits don't exist in Etoys"].
	aStream cr; tab; nextPutAll: 'instanceVariableNames: '; store: self instanceVariablesString.
	aStream cr; tab; nextPutAll: 'instanceVariableTypes: '; store: self instanceVariableTypes.
	aStream cr; tab; nextPutAll: 'classVariableNames: '; store: self classVariablesString.
	aStream cr; tab; nextPutAll: 'poolDictionaries: '; store: self sharedPoolsString.
	aStream cr; tab; nextPutAll: 'category: '; store: (SystemOrganization categoryOfElement: self name) asString.
	superclass
		ifNil: [aStream nextPutAll: '.';
				 cr.
			aStream nextPutAll: self name.
			aStream space; nextPutAll: 'superclass: nil'].
	^ aStream contents! !

!CObject class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:53'!
definitionST80
	"Only useful for the Etoys image"
	self = CObject ifTrue: [^super definitionST80].
	^ self definition! !

!CObject class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:53'!
instanceVariableTypes
	"This method will be automaticaly generated for subclasses whenever a subclass of CObject is compiled. It returns the information of the inst vars' type. I choose to do it this way (instead of using an class inst var) to ensure the type info will be saved when we file out this class or store it using Monticello"
	^ #()! !

!CObject class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:27'!
instanceVariableTypes: anArray
"Automatically compiles a method called #instanceVariableTypes that returns the array given as a parameter"
anArray isEmpty ifTrue: [^ self].
self class compile: 'instanceVariableTypes
"This method was autogenerated. Do not change unless you *really* know what you''re doing"
^ ', anArray asString classified: #definition
! !

!CObject class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:27'!
subclass: aSymbol category: aString
	^ self subclass: aSymbol instanceVariableNames: '' classVariableNames: '' poolDictionaries: '' category: aString! !

!CObject class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:27'!
subclass: nameOfClass instanceVariableNames: instVarNames instanceVariableTypes: instVarTypes classVariableNames: classVarNames poolDictionaries: poolDictnames category: category1 
	"Overwritten to allow defining the inst vars' type	info of a class"
	^ (self
		subclass: nameOfClass
		instanceVariableNames: instVarNames
		classVariableNames: classVarNames
		poolDictionaries: poolDictnames
		category: category1)
		instanceVariableTypes: instVarTypes ! !

!CObject methodsFor: 'c-comparing' stamp: 'Richo 6/22/2011 16:23'!
= t1
<returnType: #(CBoolean)>
<c: '({0} == {1})'>! !

!CObject methodsFor: 'c-pointers' stamp: 'Richo 6/23/2011 16:44'!
address
"	<returnType: #(*)>"
	<c: '&{0}'>! !

!CObject methodsFor: 'c-pointers' stamp: 'Richo 6/23/2011 16:43'!
pointer
	<c: '*{0}'>! !

!CObject methodsFor: 'c-comparing' stamp: 'Richo 6/22/2011 16:23'!
~= t1
<returnType: #(CBoolean)>
<c: '({0} !!= {1})'>! !

CObject subclass: #CPrimitiveType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

CPrimitiveType subclass: #CBlock
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CBlock class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	"Blocks are not *really* primitive types, but they're certainly not regular objects. For instance, you can't add inst vars to them. This method should not be implemented and it should'n be called as well."
	self shouldNotImplement! !

!CBlock methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:28'!
doWhileTrue: aBlock
<c: 'do\{{0}\}while({1})'>! !

!CBlock methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
repeat
<c: 'while(true)\{{0}\}'>! !

!CBlock methodsFor: 'c-controlling' stamp: 'Richo 6/21/2011 19:03'!
whileTrue
<c: 'while({0})'>! !

!CBlock methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
whileTrue: aBlock
<c: 'while({0})\{{1}\}'>! !

CPrimitiveType subclass: #CBoolean
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CBoolean class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^'bool'! !

!CBoolean methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
and: aBlock
<c: '({0} && {1})'>! !

!CBoolean methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
ifFalse: falseBlock
<c: 'if(!!({0}))\{{1}\}'>! !

!CBoolean methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
ifFalse: falseBlock ifTrue: trueBlock
<c: 'if({0})\{{2}\}else\{{1}\}'>! !

!CBoolean methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
ifTrue: aBlock
<c: 'if({0})\{{1}\}'>! !

!CBoolean methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
ifTrue: trueBlock ifFalse: falseBlock
<c: 'if({0})\{{1}\}else\{{2}\}'>! !

!CBoolean methodsFor: 'c-logical operations' stamp: 'Richo 6/20/2011 17:29'!
not
<c: '!!({0})'>! !

!CBoolean methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
or: aBlock
<c: '({0} || {1})'>! !

CPrimitiveType subclass: #CNumber
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

CNumber subclass: #CFloat
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CFloat class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^'float'! !

CNumber subclass: #CLong
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CLong class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^'long'! !

!CNumber methodsFor: 'c-arithmetic' stamp: 'Richo 6/20/2011 17:29'!
* aNumber
<c: '({0} * {1})'>! !

!CNumber methodsFor: 'c-arithmetic' stamp: 'Richo 6/20/2011 17:29'!
+ aNumber
<c: '({0} + {1})'>! !

!CNumber methodsFor: 'c-arithmetic' stamp: 'Richo 6/20/2011 17:29'!
- aNumber
<c: '({0} - {1})'>! !

!CNumber methodsFor: 'c-arithmetic' stamp: 'Richo 6/20/2011 17:29'!
/ aNumber
<c: '({0} / {1})'>! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
// aNumber
<returnType: #(CLong)> <argTypes: #(CLong)>
^ (self / aNumber) floor! !

!CNumber methodsFor: 'c-comparing' stamp: 'Richo 6/22/2011 16:23'!
< aNumber
<returnType: #(CBoolean)>
<c: '({0} < {1})'>! !

!CNumber methodsFor: 'c-comparing' stamp: 'Richo 6/22/2011 16:23'!
<= aNumber
<returnType: #(CBoolean)>
<c: '({0} <= {1})'>! !

!CNumber methodsFor: 'c-comparing' stamp: 'Richo 6/22/2011 16:23'!
> aNumber
<returnType: #(CBoolean)>
<c: '({0} > {1})'>! !

!CNumber methodsFor: 'c-comparing' stamp: 'Richo 6/22/2011 16:23'!
>= aNumber
<returnType: #(CBoolean)>
<c: '({0} >= {1})'>! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
\\ aNumber
<returnType: #(CLong)> <argTypes: #(CLong)>
^ self - (self / aNumber * aNumber)! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/20/2011 17:30'!
abs
<c: 'abs({0})'>! !

!CNumber methodsFor: 'c-truncation and round off' stamp: 'Richo 6/20/2011 17:37'!
atRandom
<c: '(Random({0}) + 1)'>! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
floor
<returnType: #(CLong)> <argTypes: #(CLong)>
^ self! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
isDivisibleBy: aNumber
<returnType: #(CBoolean)> <argTypes: #(CLong)>
aNumber = 0 ifTrue: [^false].
self \\ aNumber = 0
	ifTrue: [^ true]
	ifFalse: [^ false]! !

!CNumber methodsFor: 'c-arithmetic' stamp: 'Richo 6/20/2011 17:30'!
negated
<c: '({0} * -1)'>! !

!CNumber methodsFor: 'c-truncation and round off' stamp: 'Richo 6/20/2011 17:37'!
random
<c: '(Random({0}) + 1)'>! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
rounded
<returnType: #(CLong)>
self >= 0 
	ifTrue: [^ (self + 0.5) truncated]
	ifFalse: [^ (self - 0.5) truncated]! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/20/2011 17:36'!
sign
<c: 'sign({0})'>! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
sin
	<returnType: #(CFloat)>
	<tempTypes: #(CFloat CFloat CFloat CFloat)>
	| sum delta self2 i |
	sum := self.
	delta := self.
	self2 := 0 - (self * self).
	i := 2.
	[delta > 0.000000000001]	 whileTrue: [
		"once"
		delta := (delta * self2) / (i * (i + 1)).
		i := i + 2.
		sum := sum + delta.
		"twice"
		delta := (delta * self2) / (i * (i + 1)).
		i := i + 2.
		sum := sum + delta].
	^ sum! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/20/2011 17:36'!
sqrt
<c: 'sqrt({0})'>! !

!CNumber methodsFor: 'c-enumerating' stamp: 'Richo 6/20/2011 17:37'!
timesRepeat: aBlock
<c: 'repeat({0})\{{1}\}'>! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
truncated
<returnType: #(CLong)> <argTypes: #(CLong)>
^ self! !

CNumber subclass: #CUnsignedLong
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CUnsignedLong class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^ 'unsigned long'! !

!CPrimitiveType class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	self subclassResponsibility ! !

!CPrimitiveType class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:27'!
definition
	"Answer a String that defines the receiver."
	"Only shows #subclass:category:"
	| aStream path |
	aStream := WriteStream
				on: (String new: 300).
	superclass == nil
		ifTrue: [aStream nextPutAll: 'ProtoObject']
		ifFalse: [path := ''.
			self environment
				scopeFor: superclass name
				from: nil
				envtAndPathIfFound: [:envt :remotePath | path := remotePath].
			aStream nextPutAll: path , superclass name].
	aStream nextPutAll: self kindOfSubclass;
		 store: self name.
	aStream cr; tab; nextPutAll: 'category: '; store: (SystemOrganization categoryOfElement: self name) asString.
	superclass
		ifNil: [aStream nextPutAll: '.';
				 cr.
			aStream nextPutAll: self name.
			aStream space; nextPutAll: 'superclass: nil'].
	^ aStream contents! !

!CPrimitiveType class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:27'!
definitionST80
	"Only useful for the Etoys image"
	^ self definition! !

!CPrimitiveType class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:27'!
subclass: aSymbol instanceVariableNames: instVars classVariableNames: classVars poolDictionaries: poolDicts category: aString
	"Ignore instance variables, even if the user tries to force the superclass to add them"
	^ super subclass: aSymbol instanceVariableNames: '' classVariableNames: '' poolDictionaries: '' category: aString! !

CPrimitiveType subclass: #CString
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CString class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^'string'! !

!CString methodsFor: 'nxc-concatenation' stamp: 'Richo 6/20/2011 17:37'!
, aNXString
<c: 'StrCat({0}, {1})'>! !

CPrimitiveType subclass: #CVoid
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CVoid class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^ 'void'! !

CObject subclass: #CProgram
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

CProgram subclass: #CHelloWorld
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CHelloWorld methodsFor: 'as yet unclassified' stamp: 'Richo 6/28/2011 10:47'!
main
	self printf: 'Hello world\n';
		system: ('PAUSE')! !

!CProgram class methodsFor: 'translating' stamp: 'Richo 6/20/2011 21:25'!
translatedToC
	"Useful method to easily translate a program to c. By default the class name is used as program name"
	^ self translatedToC: self name! !

!CProgram class methodsFor: 'translating' stamp: 'Richo 6/20/2011 21:25'!
translatedToC: programName 
	"Useful method to easily translate a program to c"
	^ CProgramTranslator translate: self asProgramNamed: programName! !

!CProgram class methodsFor: 'translating' stamp: 'Richo 6/22/2011 11:12'!
translatedToCpp
	"Useful method to easily translate a program to c++. By default the class name is used as program name"
	^ self translatedToCpp: self name! !

!CProgram class methodsFor: 'translating' stamp: 'Richo 6/22/2011 11:11'!
translatedToCpp: programName 
	"Useful method to easily translate a program to c++"
	^ CppProgramTranslator translate: self asProgramNamed: programName! !

!CProgram methodsFor: 'as yet unclassified' stamp: 'Richo 6/28/2011 10:47'!
printf: aString
	<c: 'printf({1})' >! !

!CProgram methodsFor: 'as yet unclassified' stamp: 'Richo 6/23/2011 16:51'!
printf: aString value: obj
	<c: 'printf({1},{2})' >! !

!CProgram methodsFor: 'as yet unclassified' stamp: 'Richo 6/23/2011 17:54'!
system: str
	<c: 'system({1})'>! !

CProgram subclass: #CProgram1
	instanceVariableNames: 'var1 var2'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgram1 class methodsFor: 'definition' stamp: 'Richo 6/22/2011 16:28'!
instanceVariableTypes
"This method was autogenerated. Do not change unless you *really* know what you're doing"
^ #(#CLong #CLong)! !

!CProgram1 methodsFor: 'as yet unclassified' stamp: 'Richo 6/22/2011 16:23'!
main
	| var3 |
	<tempTypes: #(CLong)>
	<returnType: #(CLong)>
	var1 := 3.
	var2 := 4.
	var3 := 17.
	^ var1 + var2 - var3! !

CProgram subclass: #CProgram2
	instanceVariableNames: 'var1 var2'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgram2 class methodsFor: 'definition' stamp: 'Richo 6/22/2011 11:33'!
instanceVariableTypes
"This method was autogenerated. Do not change unless you *really* know what you're doing"
^ #(#CLong #CLong)! !

!CProgram2 methodsFor: 'as yet unclassified' stamp: 'Richo 6/22/2011 12:06'!
main
	var2 := 3.
	var1 := 4.
	self sumar: var2 y: var1 ! !

!CProgram2 methodsFor: 'as yet unclassified' stamp: 'Richo 6/22/2011 16:23'!
sumar: a y: b
	<argTypes: #(CLong CLong)>
	<returnType: #(CLong)>
	^ a + b! !

CProgram subclass: #CProgram3
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgram3 methodsFor: 'as yet unclassified' stamp: 'Richo 6/23/2011 17:54'!
main
	| p h |
	<tempTypes: #(CLong CLong*)>
	p := 45.
	h := p address.
	self printf: '%d\n' value: h pointer.
	h := h + 1.
	self printf: '%d\n' value: h.
	self system: 'PAUSE'! !

CProgram subclass: #CProgram4
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

Object subclass: #COutputBuilder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!COutputBuilder commentStamp: 'Richo 6/28/2011 12:03' prior: 0!
A COutputBuilder is responsible for constructing a set of objects that represent the output from the translators.!

!COutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 11:31'!
functionClass
	^ CTranslatedFunction! !

!COutputBuilder class methodsFor: 'building' stamp: 'Richo 6/28/2011 11:31'!
newFunctionFor: aClass node: aRBMethodNode source: aByteString 
	^ self functionClass class: aClass node: aRBMethodNode source: aByteString! !

!COutputBuilder class methodsFor: 'building' stamp: 'Richo 6/28/2011 11:24'!
newProgramNamed: aByteSymbol 
	^ self programClass named: aByteSymbol! !

!COutputBuilder class methodsFor: 'building' stamp: 'Richo 6/28/2011 11:25'!
newStructFor: aClass 
	^ self structClass for: aClass! !

!COutputBuilder class methodsFor: 'building' stamp: 'Richo 6/28/2011 11:49'!
newVariableNamed: aByteString type: aClass 
	^ self variableClass name: aByteString type: aClass! !

!COutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 11:24'!
programClass
	^ CTranslatedProgram! !

!COutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 11:25'!
structClass
	^ CStruct! !

!COutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 11:49'!
variableClass
	^ CVariable! !

COutputBuilder subclass: #CppOutputBuilder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CppOutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 11:59'!
functionClass
	^ CppTranslatedMethod! !

!CppOutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 11:59'!
programClass
	^ CppTranslatedProgram! !

!CppOutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 11:59'!
structClass
	^ CppClass! !

Object subclass: #CPointer
	instanceVariableNames: 'type'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CPointer class methodsFor: 'as yet unclassified' stamp: 'Richo 6/23/2011 17:10'!
on: aClass 
	^ (self basicNew setType: aClass) initialize! !

!CPointer methodsFor: 'as yet unclassified' stamp: 'Richo 6/23/2011 17:21'!
asCType
	^ type asCType , '*'! !

!CPointer methodsFor: 'private' stamp: 'Richo 6/23/2011 17:11'!
setType: aClass 
	type := aClass! !

!CPointer methodsFor: 'as yet unclassified' stamp: 'Richo 6/23/2011 17:21'!
ultimateCompiledMethodAt: aByteSymbol ifAbsent: aBlockClosure 
	^ type ultimateCompiledMethodAt: aByteSymbol ifAbsent: aBlockClosure ! !

Object subclass: #CPrimitiveRule
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CPrimitiveRule commentStamp: 'Richo 6/28/2011 10:53' prior: 0!
A CPrimitiveRule is is a very simple way of defining a rule to translate a smalltalk message to a c primitive. Every function (such as "printf" or "system"), every operator (such as "+" or "=="), and every control structure (such as "if", "while", or "repeat") is considered to be a c primitive. To define a method as a c primitive you should use the #c: pragma (look at the senders of #c: for examples).

To create a rule simply send '#as: CPrimitiveRule' to a well formed string. For example:

	'NumOut({2}, {3}, {1})' as: CPrimitiveRule.
	'({0} + {1})' as: CPrimitiveRule.
	'if({0})\{{1}\}else\{{2}\}' as: CPrimitiveRule.
	
Once a rule is created it can be applied to a message node using a translator by sending #applyTo:using:. This mehod will then replace all '{n}' in the string with the translated arguments ({0} represents the receiver).

Instance Variables
	value:		<String>

value
	- The string that defines how to apply the rule.
!

!CPrimitiveRule class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 20:59'!
newFrom: aString
	^ self on: aString! !

!CPrimitiveRule class methodsFor: 'instance creation' stamp: 'Richo 6/28/2011 12:06'!
on: aByteString 
	^ self basicNew initializeWith: aByteString! !

!CPrimitiveRule methodsFor: 'applying' stamp: 'Richo 6/20/2011 21:13'!
applyTo: aMessageNode using: aCMethodTranslator 
	"Visits the message node receiver and arguments and replaces the '{n}' in the string with the result ({0} represents the receiver). It reinitializes the source stream for every node so that we only get the result of visiting that node"
	^ self format: ({aMessageNode receiver}, aMessageNode arguments
		collectWithIndex: [:node :index | 
		(String streamContents: [:stream | 
				aCMethodTranslator withSource: stream do: [aCMethodTranslator visitNode: node].
				self removeSemicolonIfNecessaryOn: stream for: index - 1])])! !

!CPrimitiveRule methodsFor: 'applying' stamp: 'Richo 6/20/2011 20:59'!
format: aCollection
	"Zero based. It goes through value increasing by 1 the enclosed expressions and then it sends regular #format:"
	| oldStream newStream |
	oldStream := value readStream.
	newStream := WriteStream on: ''.
	[oldStream atEnd]
		whileFalse: [| next |
		next := oldStream next.
		newStream nextPut: next.
		(next = ${ and: [oldStream peekBack ~= $\])
			ifTrue: [| n | "We found an enclosed expression"
			n := oldStream upTo: $}.
			newStream nextPutAll: (n asNumber + 1) asString, '}'.
			]
	].
	^ newStream contents format: aCollection
	! !

!CPrimitiveRule methodsFor: 'initializing' stamp: 'Richo 6/28/2011 12:07'!
initializeWith: aByteString 
	self initialize.
	value := aByteString ! !

!CPrimitiveRule methodsFor: 'printing' stamp: 'Richo 6/20/2011 20:59'!
printOn: aStream
super printOn: aStream.
aStream nextPut: $(.
aStream nextPutAll: value.
aStream nextPut: $)! !

!CPrimitiveRule methodsFor: 'applying' stamp: 'Richo 6/20/2011 20:59'!
removeSemicolonIfNecessaryOn: source for: argIndex
	"Depending on the position of the argument in the rule string, it can be necessary to remove the semicolon after the statement. For now, we only leave the semicolon if the token is surrounded by curly brackets {}.
	For instance, in the rule below
	
	'while({0})\{{1}\}'
	
	The node replacing the {0} token will have its semicolon removed, whereas the {1} won't. Of course, most of the time it won't be necessary to remove any semicolon (if {0} is a literal or a variable, for instance)."
	| index newContents oldContents rule token |
	oldContents := source contents.
	rule := value withBlanksCondensed.
	
	"Do nothing if the source does not end with a semicolon"
	(oldContents size > 2 and: [oldContents endsWith: ';', String cr]) ifFalse: [^self].

	token := '{', argIndex asString ,'}'.
	index := rule findString: token.

	"Do nothing if the token is not in the rule"
	index = 0 ifTrue: [^self].

	"Do nothing if the token is not surrounded by curly brackets"
	((rule at: index - 2 ifAbsent: ['']) asString ,
	(rule at: index - 1 ifAbsent: ['']) asString ,
	(rule at: index + token size ifAbsent: ['']) asString ,
	(rule at: index + token size + 1 ifAbsent: ['']) asString) = '\{\}'
		ifTrue: [^self].

	"Okay, you can now safely remove the semicolon"
	newContents := oldContents allButLast: 2.
	source reset; nextPutAll: newContents.! !

Object subclass: #CProgramTranslator
	instanceVariableNames: 'program'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CProgramTranslator commentStamp: 'Richo 6/28/2011 10:51' prior: 0!
A CProgramTranslator is responsible for translating subclasses of CProgram from smalltalk to C. It's public protocol is very simple, just use #translate:asProgramNamed:, which returns an instance of CTranslatedProgram.
An even easier way of translating programs is sending #translatedToC to a CProgram subclass. For example:

	CHelloWorld translatedToC.
	
You can also specify a different program name:

	CHelloWorld translatedToC: 'Other program name'.

The translation is done recursively, starting from the #main method and translating every method call found from there. For more information, see CMethodTranslator comment.


Instance Variables
	program:		<CTranslatedProgram>

program
	- The program being translated.
!

!CProgramTranslator class methodsFor: 'output builder' stamp: 'Richo 6/28/2011 11:23'!
builder
	^ COutputBuilder! !

!CProgramTranslator class methodsFor: 'constants' stamp: 'Richo 6/20/2011 20:59'!
defaultMainSelector
"Returns the name of the function to replace the actual main of our program. Which, by necessity it's different from the main task"
	^ #'_main_'! !

!CProgramTranslator class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 20:59'!
new
	self shouldNotImplement! !

!CProgramTranslator class methodsFor: 'instance creation' stamp: 'Richo 6/28/2011 11:33'!
translate: aCProgram asProgramNamed: programName
	^ (self basicNew initializeWithProgram: (self builder newProgramNamed: programName))
			translate: aCProgram! !

!CProgramTranslator methodsFor: 'output builder' stamp: 'Richo 6/28/2011 11:25'!
builder
	^ self class builder! !

!CProgramTranslator methodsFor: 'initializing' stamp: 'Richo 6/28/2011 11:34'!
initializeWithProgram: aCTranslatedProgram 
	program := aCTranslatedProgram.
	self initialize! !

!CProgramTranslator methodsFor: 'accessing' stamp: 'Richo 6/21/2011 15:33'!
methodTranslator
	"Override if you want to define a specific method translator"
	^ CMethodTranslator! !

!CProgramTranslator methodsFor: 'translating' stamp: 'Richo 6/28/2011 11:25'!
translate: aCProgram
	program addStruct: (self builder newStructFor: aCProgram).
	[self translateMethod: #main of: aCProgram]
		on: CInstVarToImplement do: [:ex |
			(program addStruct: (self builder newStructFor: ex ownerClass))
				addVar: (self builder newVariableNamed: ex varName type: ex varType).
			ex resume].
	^ program! !

!CProgramTranslator methodsFor: 'translating' stamp: 'Richo 6/28/2011 11:27'!
translateMethod: selector of: class
	"Recursively translates this method and all methods called from it"
	[program addMethod: (self methodTranslator translate: selector of: class with: self builder)]
		on: CMethodToImplement
		do: [:exception | 
			self translateMethod: exception methodSelector of: exception methodClass.
			exception resume]! !

CProgramTranslator subclass: #CppProgramTranslator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CppProgramTranslator class methodsFor: 'output builder' stamp: 'Richo 6/28/2011 12:01'!
builder
	^ CppOutputBuilder ! !

!CppProgramTranslator methodsFor: 'as yet unclassified' stamp: 'Richo 6/22/2011 11:09'!
methodTranslator
	"Override if you want to define a specific method translator"
	^ CppMethodTranslator! !

!CppProgramTranslator methodsFor: 'as yet unclassified' stamp: 'Richo 6/28/2011 11:28'!
translateMethod: selector of: class
	"Recursively translates this method and all methods called from it"
	[program
		addStruct: (self builder newStructFor: class); "Just to ensure we have a struct for this class"
		addMethod: (self methodTranslator translate: selector of: class with: self builder)]
		on: CMethodToImplement
		do: [:exception | 
			self translateMethod: exception methodSelector of: exception methodClass.
			exception resume]! !

Object subclass: #CStruct
	instanceVariableNames: 'ownerClass variables'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CStruct class methodsFor: 'instance creation' stamp: 'Richo 6/28/2011 12:07'!
for: aClass 
	^ self basicNew initializeWith: aClass! !

!CStruct methodsFor: 'comparing' stamp: 'Richo 6/28/2011 11:55'!
= anObject
	(anObject isKindOf: self class) ifFalse: [^ false].
	^ anObject ownerClass = ownerClass and: [anObject variables = variables]! !

!CStruct methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:50'!
addVar: aCInstVar 
	variables add: aCInstVar ! !

!CStruct methodsFor: 'comparing' stamp: 'Richo 6/28/2011 11:52'!
hash
	^ ownerClass hash hashMultiply bitXor: variables hash! !

!CStruct methodsFor: 'initialize-release' stamp: 'Richo 6/28/2011 12:09'!
initializeWith: aClass
	self initialize.
	ownerClass := aClass.
	variables := Set new.! !

!CStruct methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
ownerClass
	^ ownerClass ! !

!CStruct methodsFor: 'accessing' stamp: 'Richo 6/21/2011 15:21'!
source
	^ String streamContents: [:source |
		source nextPutAll: 'typedef struct {'; cr.
		variables do: [:var |
			source nextPutAll: var source; cr].
		source nextPutAll: '} ', ownerClass asCType,';'
		]! !

!CStruct methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
variables
	^ variables ! !

CStruct subclass: #CppClass
	instanceVariableNames: 'methods'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CppClass methodsFor: 'accessing' stamp: 'Richo 6/22/2011 11:48'!
addMethod: aCppTranslatedMethod 
	methods add: aCppTranslatedMethod.! !

!CppClass methodsFor: 'initialize-release' stamp: 'Richo 6/22/2011 11:48'!
initialize
	super initialize.
	methods := Set new.! !

!CppClass methodsFor: 'accessing' stamp: 'Richo 6/28/2011 10:57'!
source
	^ String streamContents: [:source |
		source nextPutAll: 'class ', ownerClass asCType, ' {'; cr.
		source nextPutAll: 'private:'; cr.
		variables do: [:var |
			source tab; nextPutAll: var source; cr].
		source nextPutAll: 'public:'; cr.
		methods do: [:method |
			source tab; nextPutAll: method methodSignature; cr.
		].
		source nextPutAll: '};']! !

Object subclass: #CTranslatedFunction
	instanceVariableNames: 'methodClass methodNode methodSource'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CTranslatedFunction class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 20:49'!
class: aClass node: aMethodNode source: aByteString 
	^ (self basicNew setClass: aClass node: aMethodNode source: aByteString) initialize! !

!CTranslatedFunction methodsFor: 'comparing' stamp: 'Richo 6/20/2011 20:51'!
= anObject
(anObject isKindOf: self class) ifFalse: [^ false].
^ self source = anObject source! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 6/20/2011 21:00'!
argumentTypes
	^ {self selfType}, (methodNode arguments collect: [:each | CTypeCollector findTypeOf: each in: methodClass])! !

!CTranslatedFunction methodsFor: 'comparing' stamp: 'Richo 6/20/2011 20:49'!
hash
	^ self source hash hashMultiply ! !

!CTranslatedFunction methodsFor: 'testing' stamp: 'Richo 6/28/2011 12:10'!
isMain
	^ self methodName = 'main'! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:54'!
methodName
	^ methodNode selector asCIdentifier ! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 6/28/2011 10:56'!
methodSignature
	^ (methodSource upTo: Character cr), ';'! !

!CTranslatedFunction methodsFor: 'printing' stamp: 'Richo 6/20/2011 20:49'!
printOn: aStream
	aStream nextPutAll: methodSource! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 6/20/2011 21:00'!
returnType
	^ CTypeCollector findTypeOf: methodNode in: methodClass! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
selfType
	^ methodClass! !

!CTranslatedFunction methodsFor: 'private' stamp: 'Richo 6/20/2011 20:49'!
setClass: aClass node: aMethodNode source: aByteString
	methodClass := aClass.
	methodNode := aMethodNode.
	methodSource := aByteString! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
source
	^ methodSource ! !

CTranslatedFunction subclass: #CppTranslatedMethod
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CppTranslatedMethod methodsFor: 'accessing' stamp: 'Richo 6/28/2011 11:22'!
belongsTo: aClass
	^ methodClass = aClass! !

!CppTranslatedMethod methodsFor: 'accessing' stamp: 'Richo 6/28/2011 12:10'!
methodSignature
	^ String streamContents: [:stream | | cont name |
		name := self isMain ifTrue: [CProgramTranslator defaultMainSelector] ifFalse: [self methodName].
		stream nextPutAll: self returnType asCType , ' ', name , '('.
		cont := 1.
		methodNode arguments do: [:var |
			cont := cont + 1.
			stream nextPutAll: (self argumentTypes at: cont) asCType , ' ',  var name.
			] separatedBy: [stream nextPutAll: ', '].
		stream nextPutAll: ');'
		]! !

Object subclass: #CTranslatedProgram
	instanceVariableNames: 'structs methods programName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CTranslatedProgram class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 20:49'!
named: programName
	^ (self basicNew setName: programName) initialize! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:54'!
addMethod: aCTranslatedMethod 
	methods add: aCTranslatedMethod ! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/28/2011 11:55'!
addStruct: aCStruct
	"Adds a struct for the given class only if a struct is not already added. Returns the struct"
	^ structs
		detect: [:each | each ownerClass = aCStruct ownerClass]
		ifNone: [structs add: aCStruct]! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/28/2011 10:58'!
header
	"Store program's name"
	| source |
	source := WriteStream on: ''.
	source nextPutAll: '/* ', self headerFileName ,' */'; cr; cr.

	"Store structs definitions"
	structs do: [:struct |
		source nextPutAll: struct source; cr.
	].
	source cr.

	"Store methods' signatures"
	source nextPutAll: self mainSignature; cr.
	methods do: [:method |
			source nextPutAll: method methodSignature; cr.
		].
	source cr.

	^ source contents! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/28/2011 10:51'!
headerFileName
	^ programName, '.h'! !

!CTranslatedProgram methodsFor: 'initialize-release' stamp: 'Richo 6/20/2011 20:49'!
initialize
	super initialize.
	structs := Set new.
	methods := Set new! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/28/2011 12:12'!
mainImplementation
	^'int main()
\{
{1} program;
{2}(program);
return 0;
\}' format: {
	(methods detect: [:m | m isMain]) selfType asCType.
	CProgramTranslator defaultMainSelector}

! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/28/2011 10:57'!
mainSignature
	^ (self mainImplementation upTo: Character cr) , ';'! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
methods
	^ methods! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
name
	^ programName ! !

!CTranslatedProgram methodsFor: 'printing' stamp: 'Richo 6/21/2011 17:21'!
printOn: aStream
	aStream
		nextPutAll: self header; cr;
		nextPutAll: '=================='; cr;
		nextPutAll: self source! !

!CTranslatedProgram methodsFor: 'fileOut' stamp: 'Richo 6/21/2011 17:36'!
saveInDirectory: aDirectory
	aDirectory
		forceNewFileNamed: self headerFileName do: [:file | file nextPutAll: self header];
		forceNewFileNamed: self sourceFileName do: [:file | file nextPutAll: self source]! !

!CTranslatedProgram methodsFor: 'private' stamp: 'Richo 6/28/2011 10:51'!
setName: aString
	programName := aString asCIdentifier! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/21/2011 17:37'!
source
	"Store program's name"
	| source |
	source := WriteStream on: ''.
	source nextPutAll: '/* ', self sourceFileName ,' */'; cr; cr.

	"Store header include"
	source nextPutAll: '#include "', self headerFileName, '"'; cr; cr.
	
	"Store methods implementation"
	source nextPutAll: self mainImplementation; cr; cr.
	methods do: [:method |
			source nextPutAll: method source]
			separatedBy: [source cr; cr].

	^ source contents! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/28/2011 10:51'!
sourceFileName
	^ programName , '.c'! !

CTranslatedProgram subclass: #CppTranslatedProgram
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CppTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/28/2011 11:21'!
addMethod: aCppTranslatedMethod
	(structs detect: [:each | aCppTranslatedMethod belongsTo: each ownerClass])
		addMethod: aCppTranslatedMethod.
	methods add: aCppTranslatedMethod! !

!CppTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/22/2011 11:58'!
header
	"Store program's name"
	| source |
	source := WriteStream on: ''.
	source nextPutAll: '/* ', self headerFileName ,' */'; cr; cr.

	"Store structs definitions"
	structs do: [:struct |
		source nextPutAll: struct source; cr.
	].
	source cr.
	
	^ source contents! !

!CppTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/28/2011 12:12'!
mainImplementation
	^'int main()
\{
{1} program;
program.{2}();
return 0;
\}' format: {
	(methods detect: [:m | m isMain]) selfType asCType.
	CProgramTranslator defaultMainSelector}

! !

!CppTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/28/2011 10:51'!
sourceFileName 
	^ programName, '.cpp'! !

Object subclass: #CVariable
	instanceVariableNames: 'varName varType'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CVariable class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 20:49'!
name: varName type: varType
	^ (self basicNew setName: varName type: varType) initialize! !

!CVariable methodsFor: 'comparing' stamp: 'Richo 6/20/2011 20:51'!
= anObject
(anObject isKindOf: self class) ifFalse: [^ false].
^ self name = anObject name and: [self type = anObject type]! !

!CVariable methodsFor: 'comparing' stamp: 'Richo 6/20/2011 20:49'!
hash
^ (self name hash hashMultiply bitXor: self type hash) hashMultiply ! !

!CVariable methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
name
^ varName ! !

!CVariable methodsFor: 'private' stamp: 'Richo 6/20/2011 20:50'!
setName: aString type: aClass
	varName := aString.
	varType := aClass ! !

!CVariable methodsFor: 'accessing' stamp: 'Richo 6/21/2011 15:20'!
source
	^ self type asCType, ' ', self name, ';'! !

!CVariable methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
type
^ varType! !
